#import "Cedar.h"
#import "SimpleIncrementer.h"
#import "StubbedMethod.h"
#import "CedarDoubleImpl.h"
#import "FooSuperclass.h"

SHARED_EXAMPLE_GROUPS_BEGIN(CedarDoubleSharedExamples)

using namespace Cedar::Matchers;
using namespace Cedar::Doubles;
using namespace Cedar::Doubles::Arguments;

sharedExamplesFor(@"a Cedar double", ^(NSDictionary *sharedContext) {
    __block id<CedarDouble, SimpleIncrementer> myDouble;

    beforeEach(^{
        myDouble = [sharedContext objectForKey:@"double"];
    });

    describe(@"sent_messages", ^{
        beforeEach(^{
            myDouble stub_method("value");
            [myDouble value];
        });

        it(@"should have one recording per message sent", ^{
            [[myDouble sent_messages] count] should equal(1);
        });
    });

    describe(@"sent_messages_with_selector", ^{
        beforeEach(^{
            myDouble stub_method("incrementBy:");
            myDouble stub_method("incrementByInteger:");

            [myDouble incrementByInteger:1];
            [myDouble incrementBy:2];
            [myDouble incrementByInteger:3];
            [myDouble incrementBy:4];
            [myDouble incrementByInteger:5];
        });

        it(@"should return all invocations for messages sent that match the given selector", ^{
            NSArray *sentMessages = [myDouble sent_messages_with_selector:@selector(incrementBy:)];
            sentMessages.count should equal(2);

            NSInvocation *firstInvocation = sentMessages.firstObject;
            firstInvocation.selector should equal(@selector(incrementBy:));
            NSUInteger firstIncrement;
            [firstInvocation getArgument:&firstIncrement atIndex:2];
            firstIncrement should equal(2);

            NSInvocation *secondInvocation = sentMessages.lastObject;
            secondInvocation.selector should equal(@selector(incrementBy:));
            NSUInteger secondIncrement;
            [secondInvocation getArgument:&secondIncrement atIndex:2];
            secondIncrement should equal(4);
        });
    });

    context(@"where the expected type is a char * and a char * is passed", ^{
        it(@"should just work", ^{
            ^{ myDouble stub_method("methodWithCString:").with("hello"); } should_not raise_exception;
        });
    });

    describe(@"reset_sent_messages", ^{
        beforeEach(^{
            myDouble stub_method("value");
            [myDouble value];
            myDouble should have_received("value");

            [myDouble reset_sent_messages];
        });

        it(@"should remove any previously recorded invocations", ^{
            myDouble should_not have_received("value");
        });
    });

    context(@"when recording an invocation", ^{
        it(@"should release the invocations retaining the double in afterEach", ^{
            static NSInteger numInvocations = 2;
            myDouble stub_method("value");

            NSUInteger doubleRetainCount = myDouble.retainCount;

            // spies are allowed to increment the retain count of the double by 1
            // but should hand the retain over to the autorelease pool
            @autoreleasepool {
                for (NSInteger invocation = 0; invocation < numInvocations; ++invocation) {
                    [myDouble value];
                }
            }

            myDouble.retainCount should equal(doubleRetainCount + numInvocations);

            [CedarDoubleImpl class] should conform_to(@protocol(CDRHooks));
            [CedarDoubleImpl afterEach];

            myDouble.retainCount should equal(doubleRetainCount);
        });

        it(@"should exchange any block arguments with copies that can later be invoked", ^{
            __block BOOL blockWasCalled = NO;
            void *blockVariableLocationOnStack = &blockWasCalled;

            myDouble stub_method("methodWithBlock:");

            ^{
                void(^originalBlock)() = ^{
                    blockWasCalled = YES;
                };
                [myDouble methodWithBlock:originalBlock];
            }();

            NSInvocation *invocationWithBlock = [[myDouble sent_messages] lastObject];
            void(^retrievedBlock)() = nil;

            [invocationWithBlock getArgument:&retrievedBlock atIndex:2];

            //Blocks don't change memory address when copied but we can detect copying
            //by observing when it's enclosed block variables are moved to the heap.
            //See: http://www.cocoawithlove.com/2009/10/how-blocks-are-implemented-and.html
            (void *)&blockWasCalled should_not equal(blockVariableLocationOnStack);
            retrievedBlock();
            blockWasCalled should be_truthy;
        });

        it(@"should exchange any c-string arguments with copies that can later be accessed", ^{
            char *string = (char *)malloc(6);
            strcpy(string, "hello");

            myDouble stub_method("methodWithCString:");

            [myDouble methodWithCString:string];
            strcpy(string, "byeby");

            NSInvocation *invocation = [[myDouble sent_messages] lastObject];
            char *argument = NULL;
            [invocation getArgument:&argument atIndex:2];

            argument == string should_not be_truthy;
            strcmp("hello", argument) should equal(0);

            free(string);
        });
    });

    describe(@"-stub_method", ^{
        context(@"with a non-double", ^{
            it(@"should raise an exception", ^{
                NSObject *non_double = [[[NSObject alloc] init] autorelease];
                ^{ non_double stub_method("description"); } should raise_exception.with_reason([NSString stringWithFormat:@"%@ is not a double", non_double]);
            });
        });

        context(@"with a method name that the stub does not respond to", ^{
            it(@"should raise an exception", ^{
                ^{ myDouble stub_method("wibble_wobble"); } should raise_exception.with_reason([NSString stringWithFormat:@"Attempting to stub method <wibble_wobble>, which double <%@> does not respond to", myDouble]);
            });
        });

        context(@"with a method with no arguments", ^{
            context(@"when stubbed twice", ^{
                it(@"should raise an exception with a helpful message", ^{
                    myDouble stub_method("value");
                    ^{ myDouble stub_method("value"); } should raise_exception.with_reason(@"The method <value> is already stubbed with arguments () - use again() to override");
                });

                it(@"should allow the method to be stubbed again when again() is called", ^{
                    myDouble stub_method("value");
                    myDouble stub_method("value").again();
                });

                it(@"should blow up when again() is called when not previously stubbed", ^{
                    ^{ myDouble stub_method("value").again(); } should raise_exception.with_reason(@"The method <value> is not stubbed with arguments () - again() should only be used to override a stubbed method.");
                });
            });

            context(@"with no return value", ^{
                beforeEach(^{
                    myDouble stub_method("value");
                });

                it(@"should record the invocation", ^{
                    [myDouble value];
                    myDouble should have_received("value");
                });

                it(@"should return zero", ^{
                    myDouble.value should equal(0);
                });
            });

            context(@"with a return value of the correct type", ^{
                size_t someArgument = 1;

                beforeEach(^{
                    myDouble stub_method("value").and_return(someArgument);
                });

                it(@"should return the specified return value", ^{
                    myDouble.value should equal(someArgument);
                });
            });

            context(@"with a return value of a type with the wrong binary size", ^{
                int someArgument = 1;

                it(@"should raise an exception", ^{
                    ^{ myDouble stub_method("value").and_return(someArgument); } should raise_exception;
                });
            });

            context(@"with a return value of an inappropriate type", ^{
                it(@"should raise an exception", ^{
                    ^{ myDouble stub_method("value").and_return(@"foo"); } should raise_exception;
                });
            });
        });

        context(@"when the stub is instructed to raise an exception (and_raise)", ^{
            context(@"with no parameter", ^{
                beforeEach(^{
                    myDouble stub_method("increment").and_raise_exception();
                });

                it(@"should raise a generic exception", ^{
                    ^{ [myDouble increment]; } should raise_exception([NSException class]);
                });
            });

            context(@"with a specified exception", ^{
                id someException = @"that's some pig (exception)";

                beforeEach(^{
                    myDouble stub_method("increment").and_raise_exception(someException);
                });

                it(@"should raise that exception instance", ^{
                    ^{ [myDouble increment]; } should raise_exception(someException);
                });
            });
        });

        context(@"with a replacement implementation receiving the method's arguments (and_do_block)", ^{
            context(@"with a valid block that only has a return value", ^{
                __block BOOL implementation_block_called;
                size_t return_value = 123;

                beforeEach(^{
                    implementation_block_called = NO;

                    myDouble stub_method("value").and_do_block(^size_t {
                        implementation_block_called = YES;
                        return return_value;
                    });
                });

                it(@"should invoke the block", ^{
                    [myDouble value];
                    implementation_block_called should be_truthy;
                });

                it(@"should return the value returned from the block", ^{
                    [myDouble value] should equal(return_value);
                });

                context(@"when combined with an explicit return value", ^{
                    it(@"should raise an exception", ^{
                        ^{
                            myDouble stub_method("value").and_do_block(^size_t{ return 0; }).and_return(2);
                        } should raise_exception.with_reason(@"Multiple return values specified for <value>");
                    });
                });

                context(@"when added after an explicit return value", ^{
                    it(@"should raise an exception", ^{
                        ^{
                            myDouble stub_method("value").and_return(2).and_do_block(^size_t{ return 0; });
                        } should raise_exception.with_reason(@"Multiple return values specified for <value>");
                    });
                });

                context(@"when combined with an invocation block", ^{
                    it(@"should raise an exception", ^{
                        ^{
                            myDouble stub_method("value").and_do_block(^size_t{ return 0; }).and_do(^(NSInvocation *invocation) {});
                        } should raise_exception.with_reason(@"Multiple blocks specified for <value>");
                    });
                });

                context(@"when added after an invocation block", ^{
                    it(@"should raise an exception", ^{
                        ^{
                            myDouble stub_method("value").and_do(^(NSInvocation *invocation) {}).and_do_block(^size_t{ return 0; });
                        } should raise_exception.with_reason(@"Multiple blocks specified for <value>");
                    });
                });
            });

            context(@"with a valid block that has primitive integer arguments", ^{
                size_t sent_argument = 2;
                __block size_t received_argument;

                beforeEach(^{
                    received_argument = 0;

                    myDouble stub_method("incrementBy:").and_do_block(^(size_t arg) {
                        received_argument = arg;
                    });
                    [myDouble incrementBy:sent_argument];
                });

                it(@"should be passed the correct arguments", ^{
                    received_argument should equal(sent_argument);
                });
            });

            context(@"with a valid block that has primitive floating point arguments", ^{
                double sent_argument = 9876.54321;
                __block double received_argument1, received_argument2;
                double return_value = 192837.465;

                beforeEach(^{
                    received_argument1 = received_argument2 = 0;

                    myDouble stub_method("methodWithDouble1:andDouble2:").and_do_block(^double(double double1, double double2) {
                        received_argument1 = double1;
                        received_argument2 = double2;
                        return return_value;
                    });
                    [myDouble methodWithDouble1:sent_argument andDouble2:sent_argument];
                });

                it(@"should be passed the correct arguments", ^{
                    received_argument1 should equal(sent_argument);
                    received_argument2 should equal(sent_argument);
                });

                it(@"should return the value returned from the block", ^{
                    [myDouble methodWithDouble1:0 andDouble2:0] should equal(return_value);
                });
            });

            context(@"with a valid block that uses large structs", ^{
                LargeIncrementerStruct sent_argument = { 1234567, 98765432, 42, SIZE_T_MAX };
                __block LargeIncrementerStruct received_argument;
                LargeIncrementerStruct return_value = { 123, 456, 789, SIZE_T_MAX };

                beforeEach(^{
                    received_argument = {};

                    myDouble stub_method("methodWithLargeStruct1:andLargeStruct2:").and_do_block(^LargeIncrementerStruct(LargeIncrementerStruct struct1, LargeIncrementerStruct struct2) {
                        received_argument = struct2;
                        return return_value;
                    });
                    [myDouble methodWithLargeStruct1:sent_argument andLargeStruct2:sent_argument];
                });

                it(@"should be passed the correct struct arguments", ^{
                    memcmp(&received_argument, &sent_argument, sizeof(sent_argument)) should equal(0);
                });

                it(@"should return the struct value returned from the block", ^{
                    LargeIncrementerStruct returned_value = [myDouble methodWithLargeStruct1:sent_argument andLargeStruct2:sent_argument];
                    memcmp(&return_value, &returned_value, sizeof(return_value)) should equal(0);
                });
            });

            context(@"with a valid block that uses objects", ^{
                NSNumber *sent_argument = @(M_PI);
                __block NSNumber *received_argument;
                NSNumber *return_value = @(42);

                beforeEach(^{
                    received_argument = nil;

                    myDouble stub_method("methodWithNumber1:andNumber2:").and_do_block(^NSNumber *(NSNumber *num1, NSNumber *num2) {
                        received_argument = num1;
                        return return_value;
                    });
                    [myDouble methodWithNumber1:sent_argument andNumber2:sent_argument];
                });

                it(@"should be passed the correct object arguments", ^{
                    received_argument should be_same_instance_as(sent_argument);
                });

                it(@"should return the object value returned from the block", ^{
                    [myDouble methodWithNumber1:@(1) andNumber2:@(2)] should be_same_instance_as(return_value);
                });
            });

            context(@"with a valid block that takes a complex block as a parameter", ^{
                ComplexIncrementerBlock sent_argument = ^LargeIncrementerStruct(NSNumber *, LargeIncrementerStruct, id<NSCoding>){ return (LargeIncrementerStruct){}; };
                __block ComplexIncrementerBlock received_argument;

                beforeEach(^{
                    received_argument = nil;

                    myDouble stub_method("methodWithNumber:complexBlock:").and_do_block(^(NSNumber *, ComplexIncrementerBlock block) {
                        received_argument = block;
                    });
                    [myDouble methodWithNumber:@(1) complexBlock:sent_argument];
                });

                it(@"should be passed the correct block argument", ^{
                    received_argument should equal(sent_argument);
                });
            });

            context(@"with something not a block", ^{
                it(@"should raise an exception", ^{
                    ^{ myDouble stub_method("value").and_do_block(@(2)); } should raise_exception.with_reason([NSString stringWithFormat:@"Attempted to stub and do a block that isn't a block for <value>"]);
                });
            });

            context(@"with a block that does not match the method's return type", ^{
                void (^invalidBlock)(NSString *) = ^(NSString *){};
                it(@"should raise an exception", ^{
                    ^{ myDouble stub_method("methodWithString:").and_do_block(invalidBlock); } should raise_exception.with_reason(@"Invalid return type 'void' instead of 'id' for <methodWithString:>");
                });
            });

            context(@"with a block that has a different number of arguments than the method", ^{
                void (^invalidBlock)(void) = ^{};
                it(@"should raise an exception", ^{
                    ^{ myDouble stub_method("incrementBy:").and_do_block(invalidBlock); } should raise_exception.with_reason(@"Wrong number of parameters for <incrementBy:>; expected: 1; actual: 0 (not counting the special first parameter, `id self`)");
                });
            });

            context(@"with a block that has a different argument type than the method", ^{
                void (^invalidBlock)(float) = ^(float){};
                it(@"should raise an exception", ^{
                    ^{ myDouble stub_method("incrementByNumber:").and_do_block(invalidBlock); } should raise_exception.with_reason(@"Found argument type 'float', expected 'id'; argument #1 for <incrementByNumber:>");
                });
            });
        });

        context(@"with a replacement implementation receiving an invocation (and_do)", ^{
            __block BOOL replacement_invocation_called;
            __block size_t sent_argument = 2, received_argument;
            __block size_t return_value;

            beforeEach(^{
                replacement_invocation_called = NO;
                return_value = 123;
                myDouble stub_method("incrementBy:").and_do(^(NSInvocation *invocation) {
                    replacement_invocation_called = YES;
                    [invocation getArgument:&received_argument atIndex:2];
                });
                myDouble stub_method("value").and_do(^(NSInvocation *invocation) {
                    [invocation setReturnValue:&return_value];
                });

                [myDouble incrementBy:sent_argument];
            });

            it(@"should invoke the block", ^{
                replacement_invocation_called should be_truthy;
            });

            it(@"should receive the correct arguments in the invocation", ^{
                received_argument should equal(sent_argument);
            });

            it(@"should return the value provided by the NSInvocation", ^{
                [myDouble value] should equal(return_value);
            });

            context(@"when combined with an explicit return value", ^{
                it(@"should raise an exception", ^{
                    ^{
                        myDouble stub_method("value").and_do(^(NSInvocation *invocation) {}).and_return(2);
                    } should raise_exception.with_reason(@"Multiple return values specified for <value>");
                });
            });

            context(@"when added after an explicit return value", ^{
                it(@"should raise an exception", ^{
                    ^{
                        myDouble stub_method("value").and_return(2).and_do(^(NSInvocation *invocation) {});
                    } should raise_exception.with_reason(@"Multiple return values specified for <value>");
                });
            });

            context(@"when combined with an implementation block", ^{
                it(@"should raise an exception", ^{
                    ^{
                        myDouble stub_method("value").and_do(^(NSInvocation *invocation) {}).and_do_block(^size_t{ return 0; });
                    } should raise_exception.with_reason(@"Multiple blocks specified for <value>");
                });
            });

            context(@"when added after an implementation block", ^{
                it(@"should raise an exception", ^{
                    ^{
                        myDouble stub_method("value").and_do_block(^size_t{ return 0; }).and_do(^(NSInvocation *invocation) {});
                    } should raise_exception.with_reason(@"Multiple blocks specified for <value>");
                });
            });
        });

        describe(@"argument expectations", ^{
            describe(@"when stubbing the same method multiple times", ^{
                context(@"with distinctly different arguments", ^{
                    context(@"primitive arguments", ^{
                        __block BOOL firstStubWasCalled, secondStubWasCalled;
                        beforeEach(^{
                            firstStubWasCalled = secondStubWasCalled = NO;
                            myDouble stub_method("incrementByInteger:").with(1).and_do(^(NSInvocation *) {
                                firstStubWasCalled = YES;
                            });
                            myDouble stub_method("incrementByInteger:").with(3).and_do(^(NSInvocation *) {
                                secondStubWasCalled = YES;
                            });
                        });

                        it(@"should perform the stub action associated with those arguments when invoked with those arguments", ^{
                            [myDouble incrementByInteger:3];
                            secondStubWasCalled should be_truthy;
                            firstStubWasCalled should_not be_truthy;
                        });
                    });

                    context(@"object arguments", ^{
                        NSNumber *arg1 = @1;
                        NSNumber *arg2 = @2;
                        NSNumber *returnValue1 = @3;
                        NSNumber *arg3 = @88;
                        NSNumber *returnValue2 = @42;
                        __block void(^stubMethodAgainWithDifferentArugmentsBlock)();

                        beforeEach(^{
                            myDouble stub_method("methodWithNumber1:andNumber2:").with(arg1, arg2).and_return(returnValue1);
                            stubMethodAgainWithDifferentArugmentsBlock = [^{ myDouble stub_method("methodWithNumber1:andNumber2:").with(arg1, arg3).and_return(returnValue2); } copy];
                        });

                        afterEach(^{
                            [stubMethodAgainWithDifferentArugmentsBlock release];
                        });

                        it(@"should not raise an exception", ^{
                            stubMethodAgainWithDifferentArugmentsBlock should_not raise_exception;
                        });

                        context(@"when invoked", ^{
                            beforeEach(^{
                                stubMethodAgainWithDifferentArugmentsBlock();
                            });

                            it(@"should return the value associated with the corresponding arguments", ^{
                                [myDouble methodWithNumber1:arg1 andNumber2:arg2] should equal(returnValue1);
                                [myDouble methodWithNumber1:arg1 andNumber2:arg3] should equal(returnValue2);
                            });
                        });
                    });
                });

                context(@"with the same arguments", ^{
                    context(@"primitive arguments", ^{
                        NSInteger arg1 = 1;
                        it(@"should raise an exception", ^{
                            myDouble stub_method("incrementByInteger:").with(arg1);
                            ^{ myDouble stub_method("incrementByInteger:").with(arg1); } should raise_exception.with_reason(@"The method <incrementByInteger:> is already stubbed with arguments (<1>) - use again() to override");
                        });
                    });

                    context(@"object arguments", ^{
                        NSNumber *arg1 = @1;
                        NSNumber *arg2 = @2;
                        NSNumber *returnValue1 = @3;
                        NSNumber *returnValue2 = @4;

                        it(@"should raise an exception", ^{
                            myDouble stub_method("methodWithNumber1:andNumber2:").with(arg1, arg2).and_return(returnValue1);
                            ^{ myDouble stub_method("methodWithNumber1:andNumber2:").with(arg1, arg2).and_return(@91); } should raise_exception.with_reason(@"The method <methodWithNumber1:andNumber2:> is already stubbed with arguments (<1><2>) - use again() to override");
                        });

                        it(@"should allow re-stubbing with the new return value when again() is called", ^{
                            myDouble stub_method("methodWithNumber1:andNumber2:").with(arg1, arg2).and_return(returnValue1);
                            myDouble stub_method("methodWithNumber1:andNumber2:").with(arg1, arg2).again().and_return(returnValue2);

                            [myDouble methodWithNumber1:arg1 andNumber2:arg2] should equal(returnValue2);
                        });
                    });
                });

                context(@"with no arguments, then the anything argument", ^{
                    it(@"should raise an exception", ^{
                        myDouble stub_method("incrementByInteger:");
                        ^{ myDouble stub_method("incrementByInteger:").with(Arguments::anything); } should raise_exception;
                    });
                });

                context(@"with the anything argument, then with no arguments", ^{
                    it(@"should raise an exception", ^{
                        myDouble stub_method("incrementByInteger:").with(Arguments::anything);
                        ^{ myDouble stub_method("incrementByInteger:"); } should raise_exception;
                    });
                });

                context(@"with the anything argument in different positions", ^{
                    beforeEach(^{
                        myDouble stub_method("methodWithNumber1:andNumber2:").with(@1, Arguments::anything).and_return(@1);
                        myDouble stub_method("methodWithNumber1:andNumber2:").with(Arguments::anything, @2).and_return(@2);
                    });

                    it(@"should choose the correct stub to invoke based on the specific arguments", ^{
                        [myDouble methodWithNumber1:@1 andNumber2:@3] should equal(@1);
                        [myDouble methodWithNumber1:@3 andNumber2:@2] should equal(@2);
                    });

                    it(@"should choose the most recent stub if more than one stub matches ", ^{
                        [myDouble methodWithNumber1:@1 andNumber2:@2] should equal(@2);
                    });

                    it(@"should choose the overridden stub if more than one stub matches", ^{
                        myDouble stub_method("methodWithNumber1:andNumber2:").with(Arguments::anything, @2).again().and_return(@3);
                        [myDouble methodWithNumber1:@3 andNumber2:@2] should equal(@3);

                        myDouble stub_method("methodWithNumber1:andNumber2:").with(@1, Arguments::anything).again().and_return(@4);
                        [myDouble methodWithNumber1:@1 andNumber2:@3] should equal(@4);
                    });
                });

                context(@"with an 'any instance of' argument", ^{
                    __block void(^stubMethodAgainWithNoArgumentsBlock)();
                    __block void(^stubMethodAgainWithAnyInstanceOfClassArgumentBlock)();
                    __block void(^stubMethodAgainWithAnyInstanceConformingToProtocolArgumentBlock)();
                    __block FooSuperclass *specificInstance;
                    __block BarSubclass *specificBarInstance;

                    beforeEach(^{
                        specificInstance = [[[FooSuperclass alloc] init] autorelease];
                        specificBarInstance = [[[BarSubclass alloc] init] autorelease];
                        myDouble stub_method("methodWithFooSuperclass:").with(specificInstance).and_return(@"foo");
                        myDouble stub_method("methodWithFooSuperclass:").with(specificBarInstance).and_return(@"bar_specific");

                        stubMethodAgainWithNoArgumentsBlock = [^{ myDouble stub_method("methodWithFooSuperclass:").and_return(@"quux"); } copy];
                        stubMethodAgainWithAnyInstanceOfClassArgumentBlock = [^{ myDouble stub_method("methodWithFooSuperclass:").with(Arguments::any([BarSubclass class])).and_return(@"bar"); } copy];
                        stubMethodAgainWithAnyInstanceConformingToProtocolArgumentBlock = [^{ myDouble stub_method("methodWithFooSuperclass:").with(Arguments::any(@protocol(BazProtocol))).and_return(@"baz"); } copy];
                    });

                    afterEach(^{
                        [stubMethodAgainWithNoArgumentsBlock release];
                        [stubMethodAgainWithAnyInstanceOfClassArgumentBlock release];
                        [stubMethodAgainWithAnyInstanceConformingToProtocolArgumentBlock release];
                    });

                    it(@"should not raise an exception", ^{
                        stubMethodAgainWithNoArgumentsBlock should_not raise_exception;
                        stubMethodAgainWithAnyInstanceOfClassArgumentBlock should_not raise_exception;
                        stubMethodAgainWithAnyInstanceConformingToProtocolArgumentBlock should_not raise_exception;
                    });

                    context(@"with another 'any instance of class' argument", ^{
                        beforeEach(^{
                            stubMethodAgainWithAnyInstanceOfClassArgumentBlock();
                        });

                        context(@"of the same class", ^{
                            it(@"should raise an exception", ^{
                                ^{ myDouble stub_method("methodWithFooSuperclass:").with(Arguments::any([BarSubclass class])).and_return(@"bar2"); } should raise_exception.with_reason(@"The method <methodWithFooSuperclass:> is already stubbed with arguments (<Any instance of BarSubclass>) - use again() to override");
                            });
                        });

                        context(@"of a different class", ^{
                            beforeEach(^{
                                myDouble stub_method("methodWithFooSuperclass:").with(Arguments::any([QuuxSubclass class])).and_return(@"any_quux");
                            });

                            context(@"when invoked", ^{
                                it(@"should match the stub for the correct 'any instance of' class", ^{
                                    [myDouble methodWithFooSuperclass:[[[QuuxSubclass alloc] init] autorelease]] should equal(@"any_quux");
                                });
                            });
                        });
                    });

                    context(@"with another 'any instance conforming to protocol' argument", ^{
                        beforeEach(^{
                            stubMethodAgainWithAnyInstanceConformingToProtocolArgumentBlock();
                        });

                        context(@"with the same protocol", ^{
                            it(@"should raise an exception", ^{
                                ^{ myDouble stub_method("methodWithFooSuperclass:").with(Arguments::any(@protocol(BazProtocol))).and_return(@"baz2"); } should raise_exception.with_reason(@"The method <methodWithFooSuperclass:> is already stubbed with arguments (<Any instance conforming to BazProtocol>) - use again() to override");
                            });
                        });

                        context(@"with a different protocol", ^{
                            beforeEach(^{
                                myDouble stub_method("methodWithFooSuperclass:").with(Arguments::any(@protocol(NSObject))).and_return(@"any_object");
                            });

                            context(@"when invoked", ^{
                                it(@"should match the stub for the correct protocol", ^{
                                    [myDouble methodWithFooSuperclass:[[[QuuxSubclass alloc] init] autorelease]] should equal(@"any_object");
                                });
                            });
                        });
                    });

                    context(@"when invoked", ^{
                        beforeEach(^{
                            stubMethodAgainWithNoArgumentsBlock();
                            stubMethodAgainWithAnyInstanceOfClassArgumentBlock();
                            stubMethodAgainWithAnyInstanceConformingToProtocolArgumentBlock();
                        });

                        it(@"should match the stub for specific instances", ^{
                            [myDouble methodWithFooSuperclass:specificInstance] should equal(@"foo");
                            [myDouble methodWithFooSuperclass:specificBarInstance] should equal(@"bar_specific");
                        });

                        it(@"should match the stub for Arguments::any()", ^{
                            [myDouble methodWithFooSuperclass:[[[BarSubclass alloc] init] autorelease]] should equal(@"bar");
                            [myDouble methodWithFooSuperclass:[[[FooSuperclass alloc] init] autorelease]] should equal(@"quux");
                            [myDouble methodWithFooSuperclass:[[[QuuxSubclass alloc] init] autorelease]] should equal(@"baz");
                        });

                        it(@"should match the unqualified stub", ^{
                            [myDouble methodWithFooSuperclass:nil] should equal(@"quux");
                        });
                    });
                });
            });

            context(@"when specified with .with(varargs)", ^{
                NSNumber *arg1 = @1;;
                NSNumber *arg2 = @2;
                NSNumber *arg3 = @3;
                NSNumber *returnValue = @99;

                context(@"with the incorrect number of arguments", ^{
                    it(@"should raise an exception when invoked with too few arguments", ^{
                        NSString *reason = @"Wrong number of expected parameters for <methodWithNumber1:andNumber2:>; expected: 1, actual: 2";
                        ^{ myDouble stub_method("methodWithNumber1:andNumber2:").with(arg1).and_return(returnValue); } should raise_exception.with_reason(reason);
                    });

                    it(@"should raise an exception when invoked with too many arguments", ^{
                        NSString *reason = @"Wrong number of expected parameters for <methodWithNumber1:andNumber2:>; expected: 3, actual: 2";
                        ^{ myDouble stub_method("methodWithNumber1:andNumber2:").with(arg1, arg2, arg3).and_return(returnValue); } should raise_exception.with_reason(reason);
                    });
                });

                context(@"with the correct number of arguments", ^{
                    context(@"of the correct types", ^{
                        beforeEach(^{
                            myDouble stub_method("methodWithNumber1:andNumber2:").with(arg1, arg2).and_return(returnValue);
                        });

                        context(@"when invoked with a parameters of the expected value", ^{
                            it(@"should return the returnValue", ^{
                                [myDouble methodWithNumber1:arg1 andNumber2:arg2] should equal(returnValue);
                            });
                        });
                    });

                    context(@"of incorrect types", ^{
                        NSString *reason = @"Attempt to compare expected argument <10> of type 'int' with actual argument type 'id'; argument #2 for <methodWithNumber1:andNumber2:>";
                        it(@"should raise an exception", ^{
                            int invalidInt = 10;
                            ^{ myDouble stub_method("methodWithNumber1:andNumber2:").with(arg1, invalidInt); } should raise_exception.with_reason(reason);
                        });
                    });
                });
            });

            context(@"when specified with .with().and_with()", ^{
                context(@"with too few", ^{
                    unsigned int expectedIncrementValue = 1;
                    NSString *reason = [NSString stringWithFormat:@"Wrong number of expected parameters for <incrementByABit:andABitMore:>; expected: 1, actual: 2"];

                    it(@"should raise an exception", ^{
                        ^{ myDouble stub_method("incrementByABit:andABitMore:").with(expectedIncrementValue); } should raise_exception.with_reason(reason);
                    });
                });

                context(@"with too many", ^{
                    NSString *reason = [NSString stringWithFormat:@"Wrong number of expected parameters for <value>; expected: 1, actual: 0"];

                    it(@"should raise an exception", ^{
                        ^{ myDouble stub_method("value").with(@"foo"); } should raise_exception.with_reason(reason);
                    });
                });

                context(@"with the correct number", ^{
                    NSNumber *expectedBitMoreValue = @10;
                    context(@"of the correct types", ^{
                        unsigned int expectedIncrementValue = 1;

                        beforeEach(^{
                            myDouble stub_method("incrementByABit:andABitMore:").with(expectedIncrementValue).and_with(expectedBitMoreValue);
                        });

                        context(@"when invoked with a parameter of the expected value", ^{
                            it(@"should not raise an exception", ^{
                                [myDouble incrementByABit:expectedIncrementValue andABitMore:expectedBitMoreValue];
                            });
                        });
                    });

                    context(@"of incorrect types", ^{
                        context(@"where the incorrect type is an object", ^{
                            it(@"should raise an exception", ^{
                                NSString *incorrectType = @"your mom";
                                NSString *reason = [NSString stringWithFormat:@"Attempt to compare expected argument <%@> of type 'id' with actual argument type 'unsigned int'; argument #1 for <incrementByABit:andABitMore:>", incorrectType];
                                ^{ myDouble stub_method("incrementByABit:andABitMore:").with(incorrectType).and_with(expectedBitMoreValue); } should raise_exception.with_reason(reason);
                            });
                        });

                        context(@"where the incorrect type is a char *", ^{
                            it(@"should raise an exception", ^{
                                NSString *reason = [NSString stringWithFormat:@"Attempt to compare expected argument <cstring(%s)> of type 'char *' with actual argument type 'unsigned int'; argument #1 for <incrementByABit:andABitMore:>", "your mom"];
                                ^{ myDouble stub_method("incrementByABit:andABitMore:").with((char *)"your mom").and_with(expectedBitMoreValue); } should raise_exception.with_reason(reason);
                            });
                        });

                        context(@"where the incorrect type is a non-object, non-cstring pointer", ^{
                            it(@"should raise an exception", ^{
                                int anInt = 1;
                                int *ptr = &anInt;
                                NSString *reason = [NSString stringWithFormat:@"Attempt to compare expected argument <%p> of type 'int *' with actual argument type 'unsigned int'; argument #1 for <incrementByABit:andABitMore:>", ptr];
                                ^{ myDouble stub_method("incrementByABit:andABitMore:").with(ptr).and_with(expectedBitMoreValue); } should raise_exception.with_reason(reason);
                            });
                        });
                    });
                });

                context(@"with a specific value", ^{
                    size_t expectedValue = 1;

                    beforeEach(^{
                        myDouble stub_method("incrementBy:").with(expectedValue);
                    });

                    context(@"when invoked with an argument of the expected value", ^{
                        it(@"should record the invocation", ^{
                            [myDouble incrementBy:expectedValue];
                            myDouble should have_received("incrementBy:").with(expectedValue);
                        });
                    });
                });

                context(@"with nil", ^{
                    __block BOOL stubbedBehaviorWasInvoked;
                    beforeEach(^{
                        stubbedBehaviorWasInvoked = NO;
                        myDouble stub_method("incrementByNumber:").with(nil).and_do(^(NSInvocation *) {
                            stubbedBehaviorWasInvoked = YES;
                        });
                    });

                    context(@"when invoked with a nil argument", ^{
                        beforeEach(^{
                            [myDouble incrementByNumber:nil];
                        });

                        it(@"should record the invocation", ^{
                            myDouble should have_received("incrementByNumber:").with(nil);
                        });

                        it(@"should invoke the stubbed behavior", ^{
                            stubbedBehaviorWasInvoked should be_truthy;
                        });
                    });
                });

                context(@"with id-typed nil", ^{
                    __block BOOL stubbedBehaviorWasInvoked;
                    beforeEach(^{
                        stubbedBehaviorWasInvoked = NO;
                        myDouble stub_method("incrementByNumber:").with((id)nil).and_do(^(NSInvocation *) {
                            stubbedBehaviorWasInvoked = YES;
                        });
                    });

                    context(@"when invoked with a nil argument", ^{
                        beforeEach(^{
                            [myDouble incrementByNumber:nil];
                        });

                        it(@"should record the invocation", ^{
                            myDouble should have_received("incrementByNumber:").with((id)nil);
                        });

                        it(@"should invoke the stubbed behavior", ^{
                            stubbedBehaviorWasInvoked should be_truthy;
                        });
                    });
                });

                context(@"with a non-object non-char pointer, when invoked with the argument", ^{
                    __block BOOL stubbedBehaviorWasInvoked;
                    __block int *pointerToPrimitivePointer;

                    beforeEach(^{
                        int primitive = 42;
                        stubbedBehaviorWasInvoked = NO;
                        myDouble stub_method(@selector(methodWithPrimitivePointerArgument:)).with(&primitive).and_do(^(NSInvocation *) {
                            stubbedBehaviorWasInvoked = YES;
                        });

                        [myDouble methodWithPrimitivePointerArgument:&primitive];
                        pointerToPrimitivePointer = &primitive;
                    });

                    it(@"should record the invocation", ^{
                        myDouble should have_received("methodWithPrimitivePointerArgument:").with(pointerToPrimitivePointer);
                    });

                    it(@"should invoke the stubbed behavior", ^{
                        stubbedBehaviorWasInvoked should be_truthy;
                    });
                });

                context(@"with an object pointer, when invoked with the argument", ^{
                    __block BOOL stubbedBehaviorWasInvoked;
                    __block id *pointerToObjectPointer;

                    beforeEach(^{
                        NSError *error = [NSError errorWithDomain:@"ImTheTester" code:12345 userInfo:nil];

                        stubbedBehaviorWasInvoked = NO;
                        myDouble stub_method(@selector(methodWithObjectPointerArgument:)).with(&error).and_do(^(NSInvocation *) {
                            stubbedBehaviorWasInvoked = YES;
                        });

                        [myDouble methodWithObjectPointerArgument:&error];

                        myDouble should have_received("methodWithObjectPointerArgument:").with(&error);
                        pointerToObjectPointer = &error;
                    });

                    it(@"should record the invocation", ^{
                        myDouble should have_received("methodWithObjectPointerArgument:").with(pointerToObjectPointer);
                    });

                    it(@"should invoke the stubbed behavior", ^{
                        stubbedBehaviorWasInvoked should be_truthy;
                    });
                });

                context(@"with an argument specified as anything", ^{
                    NSNumber *arg1 = @3;
                    NSNumber *arg2 = @123;
                    NSNumber *returnValueWithAnythingArg = @99;
                    NSNumber *returnValueWithExplicitArgs = @100;

                    context(@"and the 'anything' method is stubbed first", ^{
                        beforeEach(^{
                            myDouble stub_method("methodWithNumber1:andNumber2:").with(anything, arg2).and_return(returnValueWithAnythingArg);
                            myDouble stub_method("methodWithNumber1:andNumber2:").with(arg1, arg2).and_return(returnValueWithExplicitArgs);
                        });

                        it(@"should allow any value for the 'anything' argument", ^{
                            [myDouble methodWithNumber1:@8 andNumber2:arg2] should equal(returnValueWithAnythingArg);
                            [myDouble methodWithNumber1:@90210 andNumber2:arg2] should equal(returnValueWithAnythingArg);
                        });

                        context(@"and stubbing the method again with anything passed for the same parameter as in the stub", ^{
                            it(@"should raise an exception", ^{
                                ^{ myDouble stub_method("methodWithNumber1:andNumber2:").with(anything).and_with(arg2).and_return(returnValueWithAnythingArg); } should raise_exception.with_reason(@"The method <methodWithNumber1:andNumber2:> is already stubbed with arguments (<anything><123>) - use again() to override");
                            });
                        });

                        context(@"when two stubs exist", ^{
                            it(@"should invoke the more specific stub", ^{
                                [myDouble methodWithNumber1:arg1 andNumber2:arg2] should equal(returnValueWithExplicitArgs);
                            });
                        });
                    });

                    context(@"and the 'anything' method is stubbed second", ^{
                        beforeEach(^{
                            myDouble stub_method("methodWithNumber1:andNumber2:").with(arg1, arg2).and_return(returnValueWithExplicitArgs);
                            myDouble stub_method("methodWithNumber1:andNumber2:").with(anything, arg2).and_return(returnValueWithAnythingArg);
                        });

                        it(@"should allow any value for the 'anything' argument", ^{
                            [myDouble methodWithNumber1:@8 andNumber2:arg2] should equal(returnValueWithAnythingArg);
                            [myDouble methodWithNumber1:@90210 andNumber2:arg2] should equal(returnValueWithAnythingArg);
                        });

                        context(@"and stubbing the method again with anything passed for the same parameter as in the stub", ^{
                            it(@"should raise an exception", ^{
                                ^{ myDouble stub_method("methodWithNumber1:andNumber2:").with(anything).and_with(arg2).and_return(returnValueWithAnythingArg); } should raise_exception.with_reason(@"The method <methodWithNumber1:andNumber2:> is already stubbed with arguments (<anything><123>) - use again() to override");
                            });
                        });

                        context(@"when two stubs exist", ^{
                            it(@"should invoke the more specific stub", ^{
                                [myDouble methodWithNumber1:arg1 andNumber2:arg2] should equal(returnValueWithExplicitArgs);
                            });
                        });
                    });
                });

                context(@"with an argument specified as any instance of a specified class", ^{
                    NSNumber *arg = @123;
                    NSNumber *returnValue = @99;

                    beforeEach(^{
                        myDouble stub_method("methodWithNumber1:andNumber2:").with(any([NSDecimalNumber class]), arg).and_return(returnValue);
                    });

                    context(@"when invoked with the correct class", ^{
                        it(@"should return the expected value", ^{
                            [myDouble methodWithNumber1:[NSDecimalNumber decimalNumberWithDecimal:[@3.14159265359 decimalValue]] andNumber2:arg] should equal(returnValue);
                        });
                    });
                });

                context(@"with an argument specified as any instance conforming to a specified protocol", ^{
                    NSNumber *arg = @123;
                    NSNumber *returnValue = @99;

                    beforeEach(^{
                        myDouble stub_method("methodWithNumber1:andNumber2:").with(any(@protocol(NSObject)), arg).and_return(returnValue);
                    });

                    context(@"when invoked with the correct class", ^{
                        it(@"should return the expected value", ^{
                            [myDouble methodWithNumber1:[NSDecimalNumber decimalNumberWithDecimal:[@3.14159265359 decimalValue]] andNumber2:arg] should equal(returnValue);
                        });
                    });
                });
            });
        });

        describe(@"return values (and_return)", ^{
            context(@"with a value of the correct type", ^{
                size_t returnValue = 1729;

                beforeEach(^{
                    myDouble stub_method("value").and_return(returnValue);
                });

                it(@"should return the expected value", ^{
                    [myDouble value] should equal(returnValue);
                });
            });

            context(@"with a value of an incorrect type", ^{
                unsigned int invalidReturnValue = 3;

                it(@"should raise an exception", ^{
                    ^{ myDouble stub_method("methodWithString:").and_return(invalidReturnValue); } should raise_exception.with_reason(@"Invalid return value type 'unsigned int' instead of 'id' for <methodWithString:>");
                });
            });

            context(@"with an arbitrary struct value", ^{
                LargeIncrementerStruct returnValue = {99, 88, 77, 66};
                beforeEach(^{
                    myDouble stub_method("methodWithLargeStruct1:andLargeStruct2:").and_return(returnValue);
                });

                it(@"should return the expected value", ^{
                    LargeIncrementerStruct returnedValue = [myDouble methodWithLargeStruct1:{} andLargeStruct2:{}];
                    memcmp(&returnValue, &returnedValue, sizeof(LargeIncrementerStruct)) should equal(0);
                });
            });

            context(@"with `nil`", ^{
                beforeEach(^{
                    myDouble stub_method("valueAsNumber").and_return(nil);
                });

                it(@"should return the expected value", ^{
                    [myDouble valueAsNumber] should be_nil;
                });
            });
        });
    });
});

SHARED_EXAMPLE_GROUPS_END
